"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TokensController = void 0;
const tslib_1 = require("tslib");
const authentication_1 = require("@bleco/authentication");
const context_1 = require("@loopback/context");
const core_1 = require("@loopback/core");
const repository_1 = require("@loopback/repository");
const rest_1 = require("@loopback/rest");
const core_2 = require("@loopx/core");
const user_core_1 = require("@loopx/user-core");
const berrors_1 = require("berrors");
const jwt = tslib_1.__importStar(require("jsonwebtoken"));
const auth_1 = require("../modules/auth");
const providers_1 = require("../providers");
const repositories_1 = require("../repositories");
const services_1 = require("../services");
let TokensController = class TokensController {
    constructor(authClientRepo, userRepo, refreshTokenRepo, revokedTokenRepo, tokenService, logger, user) {
        this.authClientRepo = authClientRepo;
        this.userRepo = userRepo;
        this.refreshTokenRepo = refreshTokenRepo;
        this.revokedTokenRepo = revokedTokenRepo;
        this.tokenService = tokenService;
        this.logger = logger;
        this.user = user;
    }
    async getToken(req, codeReader) {
        const authClient = await this.authClientRepo.findOne({
            where: {
                clientId: req.clientId,
            },
        });
        if (!authClient) {
            throw new authentication_1.AuthenticationErrors.ClientInvalid();
        }
        try {
            const code = await codeReader(req.code);
            const payload = jwt.verify(code, authClient.secret, {
                audience: req.clientId,
                issuer: process.env.JWT_ISSUER,
                algorithms: ['HS256'],
            });
            if (payload.mfa) {
                throw new authentication_1.AuthenticationErrors.UserVerificationFailed();
            }
            if (payload.userId && !(await this.userRepo.firstTimeUser(payload.userId))) {
                await this.userRepo.updateLastLogin(payload.userId);
            }
            return await this.tokenService.createJWT(payload, authClient, "ACCESS" /* LoginType.ACCESS */);
        }
        catch (error) {
            this.logger.error(error);
            if (error.name === 'TokenExpiredError') {
                throw new authentication_1.AuthenticationErrors.CodeExpired();
            }
            else if (berrors_1.BErrors.Error.prototype.isPrototypeOf.call(berrors_1.BErrors.Error.prototype, error)) {
                throw error;
            }
            else {
                throw new authentication_1.AuthenticationErrors.InvalidCredentials();
            }
        }
    }
    async exchangeToken(req) {
        const payload = await this.createTokenPayload(req);
        return this.tokenService.createJWT({
            userId: payload.refreshPayload.userId,
            externalAuthToken: payload.refreshPayload.externalAuthToken,
            externalRefreshToken: payload.refreshPayload.externalRefreshToken,
        }, payload.authClient, "RELOGIN" /* LoginType.RELOGIN */, payload.refreshPayload.tenantId);
    }
    async switchToken(req) {
        if (!req.tenantId) {
            throw new berrors_1.BErrors.BadRequest('Tenant ID is required');
        }
        if (!this.user) {
            throw new authentication_1.AuthenticationErrors.TokenInvalid();
        }
        const payload = await this.createTokenPayload(req);
        return this.tokenService.createJWT({
            user: this.user,
            externalAuthToken: payload.refreshPayload.externalAuthToken,
            externalRefreshToken: payload.refreshPayload.externalRefreshToken,
        }, payload.authClient, "RELOGIN" /* LoginType.RELOGIN */, req.tenantId);
    }
    async me() {
        if (!this.user) {
            throw new authentication_1.AuthenticationErrors.TokenInvalid();
        }
        delete this.user.deviceInfo;
        return new auth_1.AuthUser(this.user);
    }
    async createTokenPayload(req /*, token?: string */) {
        const refreshPayload = await this.refreshTokenRepo.get(req.refreshToken);
        if (!refreshPayload) {
            throw new authentication_1.AuthenticationErrors.TokenExpired();
        }
        const authClient = await this.authClientRepo.findOne({
            where: {
                clientId: refreshPayload.clientId,
            },
        });
        if (!authClient) {
            throw new authentication_1.AuthenticationErrors.ClientInvalid();
        }
        // const accessToken = token?.split(' ')[1];
        // if (!accessToken || refreshPayload.accessToken !== accessToken) {
        //   throw new AuthenticationErrors.TokenInvalid();
        // }
        await this.revokedTokenRepo.set(refreshPayload.accessToken, {
            token: refreshPayload.accessToken,
        });
        await this.refreshTokenRepo.delete(req.refreshToken);
        return {
            refreshPayload,
            authClient,
        };
    }
};
exports.TokensController = TokensController;
tslib_1.__decorate([
    (0, rest_1.post)('/auth/token', {
        description: 'Send the code received from the POST /auth/login api and get refresh token and access token (webapps)',
        responses: {
            [core_2.STATUS_CODE.OK]: {
                description: 'Token Response',
                content: {
                    [core_2.CONTENT_TYPE.JSON]: {
                        schema: { [core_2.X_TS_TYPE]: auth_1.TokenResponse },
                    },
                },
            },
            ...core_2.ErrorCodes,
        },
    }),
    tslib_1.__param(0, (0, rest_1.requestBody)()),
    tslib_1.__param(1, (0, context_1.inject)(providers_1.AuthCodeBindings.CODEREADER_PROVIDER)),
    tslib_1.__metadata("design:type", Function),
    tslib_1.__metadata("design:paramtypes", [auth_1.AuthTokenRequest, Function]),
    tslib_1.__metadata("design:returntype", Promise)
], TokensController.prototype, "getToken", null);
tslib_1.__decorate([
    (0, rest_1.post)('/auth/token-refresh', {
        description: 'Gets you a new access and refresh token once your access token is expired',
        //(both mobile and web)
        responses: {
            [core_2.STATUS_CODE.OK]: {
                description: 'New Token Response',
                content: {
                    [core_2.CONTENT_TYPE.JSON]: {
                        schema: { [core_2.X_TS_TYPE]: auth_1.TokenResponse },
                    },
                },
            },
            ...core_2.ErrorCodes,
        },
    }),
    tslib_1.__param(0, (0, rest_1.requestBody)()),
    tslib_1.__metadata("design:type", Function),
    tslib_1.__metadata("design:paramtypes", [auth_1.AuthRefreshTokenRequest]),
    tslib_1.__metadata("design:returntype", Promise)
], TokensController.prototype, "exchangeToken", null);
tslib_1.__decorate([
    (0, authentication_1.authenticate)(authentication_1.STRATEGY.BEARER, {
        passReqToCallback: true,
    }),
    (0, rest_1.post)('/auth/token-switch', {
        security: core_2.OPERATION_SECURITY_SPEC,
        description: 'To switch the access-token',
        responses: {
            [core_2.STATUS_CODE.OK]: {
                description: 'Switch access token with the tenant id provided.',
                content: {
                    [core_2.CONTENT_TYPE.JSON]: {
                        schema: { [core_2.X_TS_TYPE]: auth_1.TokenResponse },
                    },
                },
            },
            ...core_2.ErrorCodes,
        },
    }),
    tslib_1.__param(0, (0, rest_1.requestBody)()),
    tslib_1.__metadata("design:type", Function),
    tslib_1.__metadata("design:paramtypes", [auth_1.AuthRefreshTokenRequest]),
    tslib_1.__metadata("design:returntype", Promise)
], TokensController.prototype, "switchToken", null);
tslib_1.__decorate([
    (0, authentication_1.authenticate)(authentication_1.STRATEGY.BEARER, {
        passReqToCallback: true,
    }),
    (0, rest_1.get)('/auth/me', {
        security: core_2.OPERATION_SECURITY_SPEC,
        description: 'To get the user details',
        responses: {
            [core_2.STATUS_CODE.OK]: {
                description: 'User Object',
                content: {
                    [core_2.CONTENT_TYPE.JSON]: {
                        schema: { [core_2.X_TS_TYPE]: auth_1.AuthUser },
                    },
                },
            },
            ...core_2.ErrorCodes,
        },
    }),
    tslib_1.__metadata("design:type", Function),
    tslib_1.__metadata("design:paramtypes", []),
    tslib_1.__metadata("design:returntype", Promise)
], TokensController.prototype, "me", null);
exports.TokensController = TokensController = tslib_1.__decorate([
    tslib_1.__param(0, (0, repository_1.repository)(user_core_1.AuthClientRepository)),
    tslib_1.__param(1, (0, repository_1.repository)(user_core_1.UserRepository)),
    tslib_1.__param(2, (0, repository_1.repository)(repositories_1.RefreshTokenRepository)),
    tslib_1.__param(3, (0, repository_1.repository)(repositories_1.RevokedTokenRepository)),
    tslib_1.__param(4, (0, core_1.service)(services_1.TokenService)),
    tslib_1.__param(5, (0, context_1.inject)(core_2.LOGGER.LOGGER_INJECT)),
    tslib_1.__param(6, (0, context_1.inject)(authentication_1.AuthenticationBindings.CURRENT_USER)),
    tslib_1.__metadata("design:paramtypes", [user_core_1.AuthClientRepository,
        user_core_1.UserRepository,
        repositories_1.RefreshTokenRepository,
        repositories_1.RevokedTokenRepository,
        services_1.TokenService, Object, Object])
], TokensController);
//# sourceMappingURL=tokens.controller.js.map