"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.PasswordlessVerifyProvider = void 0;
const tslib_1 = require("tslib");
const authentication_1 = require("@bleco/authentication");
const context_1 = require("@loopback/context");
const core_1 = require("@loopback/core");
const repository_1 = require("@loopback/repository");
const core_2 = require("@loopx/core");
const user_core_1 = require("@loopx/user-core");
const otplib_1 = require("otplib");
const otp_1 = require("../../../otp/otp");
const providers_1 = require("../../../providers");
const repositories_1 = require("../../../repositories");
const otp_service_1 = require("../../../services/otp.service");
let PasswordlessVerifyProvider = class PasswordlessVerifyProvider {
    constructor(userRepository, otpCacheRepo, client, otpService, signup, logger) {
        this.userRepository = userRepository;
        this.otpCacheRepo = otpCacheRepo;
        this.client = client;
        this.otpService = otpService;
        this.signup = signup;
        this.logger = logger;
    }
    value() {
        return async (contact, otp) => {
            let request = otp_1.OtpRequest.from(contact);
            const contactPropName = request ? request.contactPropName : 'username';
            const user = (await this.userRepository.findOne({
                where: {
                    [contactPropName]: contact,
                },
            }));
            if (!request) {
                if (!user) {
                    this.logger.error(`Invalid OTP Request. Cannot find user with {${contactPropName}: "${contact}"}`);
                    throw new authentication_1.AuthenticationErrors.InvalidCredentials();
                }
                request = otp_1.OtpRequest.fromUser(user, contact);
            }
            request.user = user;
            // sender
            if (!otp) {
                this.logger.debug(`Sending OTP to ${contact}`);
                await this.otpService.sendOtp(request, this.client);
                // create a fake user to avoid pass the passport validation
                return (user !== null && user !== void 0 ? user : new user_core_1.User({
                    [request.contactPropName]: request.contact,
                }));
            }
            // verifier
            this.logger.debug(`Verifying OTP for ${contact}`);
            const otpCache = await this.otpCacheRepo.get(request.contact);
            if (!otpCache) {
                this.logger.error('Invalid OTP Request');
                throw new authentication_1.AuthenticationErrors.OtpExpired();
            }
            let isValid = false;
            try {
                if (otpCache.otpSecret)
                    isValid = otplib_1.totp.check(otp, otpCache.otpSecret);
            }
            catch (err) {
                this.logger.error(err);
                throw new authentication_1.AuthenticationErrors.InvalidCredentials();
            }
            if (!isValid) {
                throw new authentication_1.AuthenticationErrors.OtpInvalid();
            }
            // otp validation passed, sign up the user if not exists
            if (!user) {
                this.logger.debug(`Signing up user ${request.contact}`);
                const newUser = await this.signup({
                    [request.contactPropName]: request.contact,
                });
                if (!newUser.id) {
                    throw new authentication_1.AuthenticationErrors.InvalidCredentials();
                }
                return newUser;
            }
            return user;
        };
    }
};
exports.PasswordlessVerifyProvider = PasswordlessVerifyProvider;
exports.PasswordlessVerifyProvider = PasswordlessVerifyProvider = tslib_1.__decorate([
    tslib_1.__param(0, (0, repository_1.repository)(user_core_1.UserRepository)),
    tslib_1.__param(1, (0, repository_1.repository)(repositories_1.OtpCacheRepository)),
    tslib_1.__param(2, (0, context_1.inject)(authentication_1.AuthenticationBindings.CURRENT_CLIENT)),
    tslib_1.__param(3, (0, core_1.service)(otp_service_1.OtpService)),
    tslib_1.__param(4, (0, context_1.inject)(providers_1.SignUpBindings.LOCAL_SIGNUP_PROVIDER)),
    tslib_1.__param(5, (0, context_1.inject)(core_2.LOGGER.LOGGER_INJECT)),
    tslib_1.__metadata("design:paramtypes", [user_core_1.UserRepository,
        repositories_1.OtpCacheRepository,
        user_core_1.AuthClient,
        otp_service_1.OtpService, Function, Object])
], PasswordlessVerifyProvider);
//# sourceMappingURL=passwordless-verify.provider.js.map